{------------------------------------------------------------------------------}
{                                                                              }
{  TWin32Res v1.11 - Manipulates Windows 32 Resource file format               }
{  by Kambiz R. Khojasteh                                                      }
{                                                                              }
{  kambiz@delphiarea.com                                                       }
{  http://www.delphiarea.com                                                   }
{                                                                              }
{  The following classes are provided "AS IS" whitout any warranty of any      }
{  kind, either express or implied. Use them as your own risk.                 }
{                                                                              }
{------------------------------------------------------------------------------}

unit Win32Res;

interface

uses
  Windows, SysUtils, Classes;

const
  rfMOVEABLE    = $0010;
  rfPURE        = $0020;
  rfPRELOAD     = $0040;
  rfDISCARDABLE = $1000;

const
  VerSignature    = $FEEF04BD;
  VerStrucVersion = $00010000;

type

  EWin32ResError = class(Exception);

{ TWin32Res }

  TWin32Res = class(TStream)
  private
    function ThisTagIsEqualTo(ResTag: PChar): Boolean;
    function SeekResource(ResName, ResType: PChar): Boolean;
  protected
    function IsResource: Boolean; dynamic;
    function SeekResourceData(ResName, ResType: PChar): Boolean; dynamic;
  public
    function SizeOfResource(ResName, ResType: PChar): DWord; dynamic;
    function LoadResource(ResName, ResType: PChar): TMemoryStream; dynamic;
    procedure SaveResource(ResName, ResType: PChar; Flags: Word; Stream: TStream); dynamic;
  end;

{ TWin32ResFile }

  TWin32ResFile = class(TWin32Res)
  private
    FHandle: Integer;
  protected
    procedure SetSize(NewSize: Longint); override;
  public
    constructor Create(const FileName: String; Create: Boolean);
    destructor Destroy; override;
    function Read(var Buffer; Count: Longint): Longint; override;
    function Write(const Buffer; Count: Longint): Longint; override;
    function Seek(Offset: Longint; Origin: Word): Longint; override;
    property Handle: Integer read FHandle;
  end;

{ TWin32ResLib }

  TWin32ResLib = class(TWin32Res)
  private
    FHandle: Integer;
  public
    constructor Create(const LibName: String); 
    destructor Destroy; override;
    function Read(var Buffer; Count: Longint): Longint; override;
    function Write(const Buffer; Count: Longint): Longint; override;
    function Seek(Offset: Longint; Origin: Word): Longint; override;
    function SizeOfResource(ResName, ResType: PChar): DWord; override;
    function LoadResource(ResName, ResType: PChar): TMemoryStream; override;
    procedure SaveResource(ResName, ResType: PChar; Flags: Word; Stream: TStream); override;
    property Handle: Integer read FHandle;
  end;

{ TVerRes }

  TVersion = record
  case Integer of
    1: (Ver: Int64);
    2: (MS, LS: DWord);
    3: (Minor, Major, Build, Release: Word);
    4: (V: array[1..4] of Word);
  end;

  TVerRes = class(TObject)
  private
    FStrings: TStringList;
  public
    Language: Word;
    Codepage: Word;
    FixedInfo: VS_FIXEDFILEINFO;
    constructor Create;
    destructor Destroy; override;
    procedure Clear;
    procedure LoadFromStream(Stream: TMemoryStream);
    procedure SaveToStream(Stream: TMemoryStream);
    property Strings: TStringList read FStrings;
  end;

function SplitVersion(const Ver: String): TVersion;

implementation

resourcestring
  SInvalidResourceFile = '%s is not a Win32 resource file';
  SInvalidResourceLib  = '%s is not a valid library';
  SFileCreateError     = 'Error on creating %s';
  SFileOpenError       = 'Error on opening %s';

const
  Zero: DWORD          = 0;
  VerInfoKey           = 'VS_VERSION_INFO';
  VerVarInfoKey        = 'VarFileInfo';
  VerStringInfoKey     = 'StringFileInfo';
  VerTranslationKey    = 'Translation';

type
  TVerHeader = record
    wLength: Word;
    wValueLength: Word;
    wType: Word;
  end;

{ TWin32Res }

function TWin32Res.IsResource: Boolean;
var
  Offset: Integer;
  FileTag: array[0..31] of Byte;
begin
  Result := False;
  Position := 0;
  if Read(FileTag, SizeOf(FileTag)) = SizeOf(FileTag) then
  begin
    for Offset := Low(FileTag) to High(FileTag) do
      case Offset of
        4: if FileTag[Offset] <> SizeOf(FileTag) then Exit;
        8,9,12,13: if FileTag[Offset] <> $FF then Exit;
        else if FileTag[Offset] <> 0 then Exit;
      end;
    Result := True;
  end;
end;

function TWin32Res.ThisTagIsEqualTo(ResTag: PChar): Boolean;
var
  W: Word;
  DW: DWord;
  I: Integer;
  Name: array[0..255] of Char;
begin
  Result := False;
  Read(DW, SizeOf(DW));
  if (LoWord(DW) = $FFFF) and (HiWord(DWord(ResTag)) = 0) then
    Result := HiWord(DW) = LoWord(DWord(ResTag))
  else if (LoWord(DW) <> $FFFF) and (HiWord(DWord(ResTag)) <> 0) then
  begin
    Name[0] := Chr(LoByte(LoWord(DW)));
    I := 1;
    W := HiWord(DW);
    while W <> 0 do
    begin
      Name[I] := Chr(LoByte(W));
      Inc(I);
      Read(W, SizeOf(Word));
    end;
    Name[I] := #0;
    Result := StrIComp(Name, ResTag) = 0;
  end;
end;

function TWin32Res.SeekResource(ResName, ResType: PChar): Boolean;
var
  OldPos, Pos: LongInt;
  DataSize, HeaderSize: DWord;
begin
  Result := False;
  OldPos := Position;
  Pos := 32;
  repeat
    Position := Pos;
    Read(DataSize, SizeOf(DataSize));
    Read(HeaderSize, SizeOf(HeaderSize));
    if not (ThisTagIsEqualTo(ResType) and ThisTagIsEqualTo(ResName)) then
    begin
      if (DataSize mod 4) <> 0 then
        Inc(DataSize, 4 - (DataSize mod 4));
      Inc(Pos, HeaderSize + DataSize);
    end
    else
    begin
      Position := Pos;
      Result := True;
    end;
  until (Pos >= Size) or Result;
  if not Result then
    Position := OldPos;
end;

function TWin32Res.SeekResourceData(ResName, ResType: PChar): Boolean;
var
  DataSize, HeaderSize: DWord;
begin
  Result := False;
  if SeekResource(ResName, ResType) then
  begin
    Read(DataSize, SizeOf(DataSize));
    Read(HeaderSize, SizeOf(HeaderSize));
    Position := (DWord(Position) + HeaderSize) - (SizeOf(DataSize) + SizeOf(HeaderSize));
    Result := True;
  end;
end;

function TWin32Res.SizeOfResource(ResName, ResType: PChar): DWord;
begin
  Result := 0;
  if SeekResource(ResName, ResType) then
    Read(Result, SizeOf(Result));
end;

function TWin32Res.LoadResource(ResName, ResType: PChar): TMemoryStream;
var
  DataSize, HeaderSize: DWord;
begin
  Result := nil;
  if SeekResource(ResName, ResType) then
  begin
    Read(DataSize, SizeOf(DataSize));
    if DataSize > 0 then
    begin
      Read(HeaderSize, SizeOf(HeaderSize));
      Position := (DWord(Position) + HeaderSize) - (SizeOf(DataSize) + SizeOf(HeaderSize));
      Result := TMemoryStream.Create;
      Result.CopyFrom(Self, DataSize);
    end;
  end;
end;

procedure TWin32Res.SaveResource(ResName, ResType: PChar; Flags: Word;
  Stream: TStream);
var
  ResTypeSize, ResNameSize: Word;
  HeaderSize, DataSize: DWord;
  Padding: Word;
  W, I: Word;
  DW: DWord;
begin
  Position := Size;
  if Stream <> nil then
    DataSize := Stream.Size
  else
    DataSize := 0;
  if HiWord(DWord(ResType)) = 0 then
    ResTypeSize := SizeOf(DWord)
  else
    ResTypeSize := 2 * (StrLen(ResType) + 1);
  if HiWord(DWord(ResName)) = 0 then
    ResNameSize := SizeOf(DWord)
  else
    ResNameSize := 2 * (StrLen(ResName) + 1);
  if ((ResTypeSize + ResNameSize) mod 4) <> 0 then
    Padding := 4 - ((ResTypeSize + ResNameSize) mod 4)
  else
    Padding := 0;
  HeaderSize := 24 + ResTypeSize + ResNameSize + Padding;
  Write(DataSize, SizeOf(DataSize));        // DataSize
  Write(HeaderSize, SizeOf(HeaderSize));    // HeaderSize
  if HiWord(DWord(ResType)) = 0 then
  begin
    DW := (DWord(ResType) shl 16) or $0000FFFF;
    Write(DW, SizeOf(DW));                  // ResType
  end
  else
    for I := 0 to StrLen(ResType) do
    begin
      W := Ord(ResType[I]);
      Write(W, SizeOf(W));
    end;
  if HiWord(DWord(ResName)) = 0 then
  begin
    DW := (DWord(ResName) shl 16) or $0000FFFF;
    Write(DW, SizeOf(DW));                  // ResName
  end
  else
    for I := 0 to StrLen(ResName) do
    begin
      W := Ord(ResName[I]);
      Write(W, SizeOf(W));
    end;
  if Padding <> 0 then
    Write(Zero, Padding);
  Write(Zero, SizeOf(DWord));               // DataVersion
  Write(Flags, SizeOf(Flags));              // MemoryFlags
  Write(Zero, SizeOf(Word));                // LanguageID
  Write(Zero, SizeOf(DWord));               // Version
  Write(Zero, SizeOf(DWord));               // Characteristics
  if Stream <> nil then
  begin
     CopyFrom(Stream, 0);
     if (DataSize mod 4) <> 0 then
       Write(Zero, 4 - (DataSize mod 4));
  end;
end;

{ TWin32ResFile }

constructor TWin32ResFile.Create(const FileName: String; Create: Boolean);
begin
  if Create then
  begin
    FHandle := FileCreate(FileName);
    if FHandle > 0 then
      SaveResource(MakeIntResource(0), MakeIntResource(0), 0, nil)
    else
      EWin32ResError.CreateFmt(SFileCreateError, [FileName]);
  end
  else
  begin
    FHandle := FileOpen(FileName, fmOpenRead or fmShareDenyWrite);
    if FHandle > 0 then
    begin
      if not IsResource then
        EWin32ResError.CreateFmt(SInvalidResourceFile, [FileName]);
    end
    else
      EWin32ResError.CreateFmt(SFileOpenError, [FileName]);
  end;
  inherited Create;
end;

destructor TWin32ResFile.Destroy;
begin
  if FHandle >= 0 then FileClose(FHandle);
  inherited Destroy;
end;

function TWin32ResFile.Read(var Buffer; Count: Longint): Longint;
begin
  Result := FileRead(FHandle, Buffer, Count);
  if Result = -1 then Result := 0;
end;

function TWin32ResFile.Write(const Buffer; Count: Longint): Longint;
begin
  Result := FileWrite(FHandle, Buffer, Count);
  if Result = -1 then Result := 0;
end;

function TWin32ResFile.Seek(Offset: Longint; Origin: Word): Longint;
begin
  Result := FileSeek(FHandle, Offset, Origin);
end;

procedure TWin32ResFile.SetSize(NewSize: Longint);
begin
  Seek(NewSize, soFromBeginning);
  Win32Check(SetEndOfFile(FHandle));
end;

{ TWin32ResLib }

constructor TWin32ResLib.Create(const LibName: String);
begin
  FHandle := LoadLibrary(PChar(LibName));
  if FHandle <= 32 then
    EWin32ResError.CreateFmt(SInvalidResourceLib, [LibName]);
  inherited Create;
end;

destructor TWin32ResLib.Destroy;
begin
  FreeLibrary(FHandle);
  inherited Destroy;
end;

function TWin32ResLib.Read(var Buffer; Count: Longint): Longint;
begin
  // Nothing to do
  Result := 0;
end;

function TWin32ResLib.Write(const Buffer; Count: Longint): Longint;
begin
  // Nothing to do
  Result := 0;
end;

function TWin32ResLib.Seek(Offset: Longint; Origin: Word): Longint;
begin
  // Nothing to do
  Result := 0;
end;

function TWin32ResLib.SizeOfResource(ResName, ResType: PChar): DWord;
var
  HResInfo: HRSRC;
begin
  Result := 0;
  HResInfo := Windows.FindResource(FHandle, ResName, ResType);
  if HResInfo <> 0 then
    Result := Windows.SizeOfResource(FHandle, HResInfo);
end;

function TWin32ResLib.LoadResource(ResName, ResType: PChar): TMemoryStream;
var
  HResInfo: HRSRC;
  HGlobal: THandle;
begin
  Result := nil;
  HResInfo := Windows.FindResource(FHandle, ResName, ResType);
  if HResInfo <> 0 then
  begin
    HGlobal := Windows.LoadResource(FHandle, HResInfo);
    if HGlobal <> 0 then
      try
        Result := TMemoryStream.Create;
        Result.Write(LockResource(HGlobal)^, Windows.SizeOfResource(FHAndle, HResInfo));
      finally
        UnlockResource(HGlobal);
        FreeResource(HGlobal);
      end;
  end;
end;

procedure TWin32ResLib.SaveResource(ResName, ResType: PChar; Flags: Word; Stream: TStream);
begin
  // TWin32ResLib is read-only
end;

{ TVerRes }

constructor TVerRes.Create;
begin
  inherited Create;
  FStrings := TStringList.Create;
  Clear;
end;

destructor TVerRes.Destroy;
begin
  FStrings.Free;
  inherited Destroy;
end;

procedure TVerRes.Clear;
begin
  Language := 0;
  Codepage := 0;
  FStrings.Clear;
  FillChar(FixedInfo, SizeOf(VS_FIXEDFILEINFO), 0);
  FixedInfo.dwSignature := VerSignature;
  FixedInfo.dwStrucVersion := VerStrucVersion;
end;

procedure TVerRes.LoadFromStream(Stream: TMemoryStream);

  procedure ReadString;
  var
    W: Word;
    Ver: TVerHeader;
    CheckValue: Integer;
    StartPosition: LongInt;
    Name, Value: String;
  begin
    StartPosition := Stream.Position;
    CheckValue := Stream.Position mod 4;
    // String.wLength
    Stream.Read(Ver.wLength, SizeOf(Word));
    // String.wValueLength
    Stream.Read(Ver.wValueLength, SizeOf(Word));
    // String.wType
    Stream.Read(Ver.wType, SizeOf(Word));
    // String.szKey
    Name := '';
    repeat
      Stream.Read(W, SizeOf(Word));
      if W <> 0 then Name := Name + Chr(W);
    until W = 0;
    // String.Padding
    while (Stream.Position mod 4 <> CheckValue) and
          (Stream.Position - StartPosition < Ver.wLength) do
      Stream.Read(W, SizeOf(Word));
    // String.Children
    Value := '';
    repeat
      Stream.Read(W, SizeOf(Word));
      if W <> 0 then Value := Value + Chr(W);
    until W = 0;
    FStrings.Values[Name] := Value;
  end;

  procedure ReadStringTable;
  var
    W: Word;
    Ver: TVerHeader;
    CheckValue: Integer;
    StartPosition: LongInt;
  begin
    StartPosition := Stream.Position;
    CheckValue := Stream.Position mod 4;
    // StringTable.wLength
    Stream.Read(Ver.wLength, SizeOf(Word));
    // StringTable.wValueLength
    Stream.Read(Ver.wValueLength, SizeOf(Word));
    // StringTable.wType
    Stream.Read(Ver.wType, SizeOf(Word));
    // StringTable.szKey
    repeat
      Stream.Read(W, SizeOf(Word))
    until W = 0;
    // StringTable.Padding
    while (Stream.Position mod 4 <> CheckValue) and
          (Stream.Position - StartPosition < Ver.wLength) do
      Stream.Read(W, SizeOf(Word));
    while (Stream.Position - StartPosition < Ver.wLength) do
    begin
      // StringTable.Children
      ReadString;
      // StringTable.Padding
      while (Stream.Position mod 4 <> CheckValue) and
            (Stream.Position - StartPosition < Ver.wLength) do
        Stream.Read(W, SizeOf(Word));
    end;
  end;

  procedure ReadStringFileInfo;
  var
    W: Word;
    Ver: TVerHeader;
    CheckValue: Integer;
    StartPosition: LongInt;
  begin
    StartPosition := Stream.Position;
    CheckValue := Stream.Position mod 4;
    // StringFileInfo.wLength
    Stream.Read(Ver.wLength, SizeOf(Word));
    // StringFileInfo.wValueLength
    Stream.Read(Ver.wValueLength, SizeOf(Word));
    // StringFileInfo.wType
    Stream.Read(Ver.wType, SizeOf(Word));
    // StringFileInfo.szKey
    repeat
      Stream.Read(W, SizeOf(Word))
    until W = 0;
    // StringFileInfo.Padding
    while (Stream.Position mod 4 <> CheckValue) and
          (Stream.Position - StartPosition < Ver.wLength) do
      Stream.Read(W, SizeOf(Word));
    // StringFileInfo.Children
    while (Stream.Position - StartPosition < Ver.wLength) do
      ReadStringTable;
  end;

  procedure ReadVar;
  var
    W: Word;
    Ver: TVerHeader;
    CheckValue: Integer;
    StartPosition: LongInt;
  begin
    StartPosition := Stream.Position;
    CheckValue := Stream.Position mod 4;
    // Var.wLength
    Stream.Read(Ver.wLength, SizeOf(Word));
    // Var.wValueLength
    Stream.Read(Ver.wValueLength, SizeOf(Word));
    // Var.wType
    Stream.Read(Ver.wType, SizeOf(Word));
    // Var.szKey
    repeat
      Stream.Read(W, SizeOf(Word))
    until W = 0;
    // Var.Padding
    while (Stream.Position mod 4 <> CheckValue) and
          (Stream.Position - StartPosition < Ver.wLength) do
      Stream.Read(W, SizeOf(Word));
    // Var.Children
    while (Stream.Position - StartPosition < Ver.wLength) do
    begin
      Stream.Read(Language, SizeOf(Word));
      Stream.Read(Codepage, SizeOf(Word));
    end;
  end;

  procedure ReadVarFileInfo;
  var
    W: Word;
    Ver: TVerHeader;
    CheckValue: Integer;
    StartPosition: LongInt;
  begin
    StartPosition := Stream.Position;
    CheckValue := Stream.Position mod 4;
    // VarFileInfo.wLength
    Stream.Read(Ver.wLength, SizeOf(Word));
    // VarFileInfo.wValueLength
    Stream.Read(Ver.wValueLength, SizeOf(Word));
    // VarFileInfo.wType
    Stream.Read(Ver.wType, SizeOf(Word));
    // VarFileInfo.szKey
    repeat
      Stream.Read(W, SizeOf(Word))
    until W = 0;
    // VarFileInfo.Padding
    while (Stream.Position mod 4 <> CheckValue) and
          (Stream.Position - StartPosition < Ver.wLength) do
      Stream.Read(W, SizeOf(Word));
    // VarFileInfo.Children
    while (Stream.Position - StartPosition < Ver.wLength) do
      ReadVar;
  end;

var
  W: Word;
  Ver: TVerHeader;
  CheckValue: Integer;
  StartPosition: LongInt;
begin
  Clear;
  StartPosition := Stream.Position;
  CheckValue := Stream.Position mod 4;
  // VS_VERSION_INFO.wLength
  Stream.Read(Ver.wLength, SizeOf(Word));
  // VS_VERSION_INFO.wValueLength
  Stream.Read(Ver.wValueLength, SizeOf(Word));
  // VS_VERSION_INFO.wType
  Stream.Read(Ver.wType, SizeOf(Word));
  // VS_VERSION_INFO.szKey
  repeat
    Stream.Read(W, SizeOf(Word))
  until W = 0;
  // VS_VERSION_INFO.Padding1
  while (Stream.Position mod 4 <> CheckValue) and
        (Stream.Position - StartPosition < Ver.wLength) do
    Stream.Read(W, SizeOf(Word));
  // VS_VERSION_INFO.Value
  if Ver.wValueLength <> 0 then
    Stream.Read(FixedInfo, SizeOf(VS_FIXEDFILEINFO));
  // VS_VERSION_INFO.Padding2
  while (Stream.Position mod 4 <> CheckValue) and
        (Stream.Position - StartPosition < Ver.wLength) do
    Stream.Read(W, SizeOf(Word));
  // VS_VERSION_INFO.Children
  if Stream.Position - StartPosition < Ver.wLength then
    ReadStringFileInfo;
  if Stream.Position - StartPosition < Ver.wLength then
    ReadVarFileInfo;
end;

procedure TVerRes.SaveToStream(Stream: TMemoryStream);

  procedure WriteString(Index: Integer);
  var
    W: Word;
    Ver: TVerHeader;
    I, CheckValue: Integer;
    StartPosition: LongInt;
    Name, Value: String;
  begin
    StartPosition := Stream.Position;
    CheckValue := Stream.Position mod 4;
    Name := Copy(FStrings[Index], 1, Pos('=', FStrings[Index])-1);
    Value := FStrings.Values[Name];
    Ver.wLength := 0; // Unknown at this time
    Ver.wValueLength := Length(Value) + 1;
    Ver.wType := 1;
    // String.wLength
    Stream.Write(Ver.wLength, SizeOf(Word));
    // String.wValueLength
    Stream.Write(Ver.wValueLength, SizeOf(Word));
    // String.wType
    Stream.Write(Ver.wType, SizeOf(Word));
    // String.szKey
    for I := 1 to Length(Name) do
    begin
      W := Ord(Name[I]);
      Stream.Write(W, SizeOf(Word));
    end;
    // String.Padding
    W := 0;
    repeat
      Stream.Write(W, SizeOf(Word))
    until Stream.Position mod 4 = CheckValue;
    // String.Children
    for I := 1 to Length(Value) do
    begin
      W := Ord(Value[I]);
      Stream.Write(W, SizeOf(Word));
    end;
    W := 0;
    Stream.Write(W, SizeOf(Word));
    // Update StringTable.wLength
    Ver.wLength := Stream.Position - StartPosition;
    // Update Stream
    Stream.Position := StartPosition;
    Stream.Write(Ver.wLength, SizeOf(Word));
    Stream.Position := StartPosition + Ver.wLength;
  end;

  procedure WriteStringTable;
  var
    W: Word;
    Ver: TVerHeader;
    I, CheckValue: Integer;
    StartPosition: LongInt;
    Translation: String;
  begin
    StartPosition := Stream.Position;
    CheckValue := Stream.Position mod 4;
    Translation := IntToHex(Language, 4) + IntToHex(CodePage, 4);
    Ver.wLength := 0; // Unknown at this time
    Ver.wValueLength := 0;
    Ver.wType := 0;
    // StringTable.wLength
    Stream.Write(Ver.wLength, SizeOf(Word));
    // StringTable.wValueLength
    Stream.Write(Ver.wValueLength, SizeOf(Word));
    // StringTable.wType
    Stream.Write(Ver.wType, SizeOf(Word));
    // StringTable.szKey
    for I := 1 to Length(Translation) do
    begin
      W := Ord(Translation[I]);
      Stream.Write(W, SizeOf(Word));
    end;
    // StringTable.Padding
    W := 0;
    repeat
      Stream.Write(W, SizeOf(Word))
    until Stream.Position mod 4 = CheckValue;
    for I := 0 to FStrings.Count-1 do
    begin
      // StringTable.Children
      WriteString(I);
      // StringTable.Padding
      while Stream.Position mod 4 <> CheckValue do
        Stream.Write(W, SizeOf(Word));
    end;
    // Update StringTable.wLength
    Ver.wLength := Stream.Position - StartPosition;
    Stream.Position := StartPosition;
    Stream.Write(Ver.wLength, SizeOf(Word));
    Stream.Position := StartPosition + Ver.wLength;
  end;

  procedure WriteStringFileInfo;
  var
    W: Word;
    Ver: TVerHeader;
    I, CheckValue: Integer;
    StartPosition: LongInt;
  begin
    StartPosition := Stream.Position;
    CheckValue := Stream.Position mod 4;
    Ver.wLength := 0; // Unknown at this time
    Ver.wValueLength := 0;
    Ver.wType := 0;
    // StringFileInfo.wLength
    Stream.Write(Ver.wLength, SizeOf(Word));
    // StringFileInfo.wValueLength
    Stream.Write(Ver.wValueLength, SizeOf(Word));
    // StringFileInfo.wType
    Stream.Write(Ver.wType, SizeOf(Word));
    // StringFileInfo.szKey
    for I := 1 to Length(VerStringInfoKey) do
    begin
      W := Ord(VerStringInfoKey[I]);
      Stream.Write(W, SizeOf(Word));
    end;
    // StringFileInfo.Padding
    W := 0;
    repeat
      Stream.Write(W, SizeOf(Word))
    until Stream.Position mod 4 = CheckValue;
    // StringFileInfo.Children
    WriteStringTable;
    // Update StringFileInfo.wLength
    Ver.wLength := Stream.Position - StartPosition;
    // Update StringFileInfo.wValueLength
    Stream.Position := StartPosition;
    Stream.Write(Ver.wLength, SizeOf(Word));
    Stream.Position := StartPosition + Ver.wLength;
  end;

  procedure WriteVar;
  var
    W: Word;
    Ver: TVerHeader;
    I, CheckValue: Integer;
    StartPosition: LongInt;
  begin
    StartPosition := Stream.Position;
    CheckValue := Stream.Position mod 4;
    Ver.wLength := 0; // Unknown at this time
    Ver.wValueLength := SizeOf(Language) + SizeOf(CodePage);
    Ver.wType := 0;
    // Var.wLength
    Stream.Write(Ver.wLength, SizeOf(Word));
    // Var.wValueLength
    Stream.Write(Ver.wValueLength, SizeOf(Word));
    // Var.wType
    Stream.Write(Ver.wType, SizeOf(Word));
    // Var.szKey
    for I := 1 to Length(VerTranslationKey) do
    begin
      W := Ord(VerTranslationKey[I]);
      Stream.Write(W, SizeOf(Word));
    end;
    // Var.Padding
    W := 0;
    repeat
      Stream.Write(W, SizeOf(Word))
    until Stream.Position mod 4 = CheckValue;
    // Var.Children
    Stream.Write(Language, SizeOf(Word));
    Stream.Write(Codepage, SizeOf(Word));
    // Update Var.wLength
    Ver.wLength := Stream.Position - StartPosition;
    Stream.Position := StartPosition;
    Stream.Write(Ver.wLength, SizeOf(Word));
    Stream.Position := StartPosition + Ver.wLength;
  end;

  procedure WriteVarFileInfo;
  var
    W: Word;
    Ver: TVerHeader;
    I, CheckValue: Integer;
    StartPosition: LongInt;
  begin
    StartPosition := Stream.Position;
    CheckValue := Stream.Position mod 4;
    Ver.wLength := 0; // Unknown at this time
    Ver.wValueLength := 0;
    Ver.wType := 0;
    // VarFileInfo.wLength
    Stream.Write(Ver.wLength, SizeOf(Word));
    // VarFileInfo.wValueLength
    Stream.Write(Ver.wValueLength, SizeOf(Word));
    // VarFileInfo.wType
    Stream.Write(Ver.wType, SizeOf(Word));
    // VarFileInfo.szKey
    for I := 1 to Length(VerVarInfoKey) do
    begin
      W := Ord(VerVarInfoKey[I]);
      Stream.Write(W, SizeOf(Word));
    end;
    // VarFileInfo.Padding
    W := 0;
    repeat
      Stream.Write(W, SizeOf(Word))
    until Stream.Position mod 4 = CheckValue;
    // VarFileInfo.Children
    WriteVar;
    // Update VarFileInfo.wLength
    Ver.wLength := Stream.Position - StartPosition;
    Stream.Position := StartPosition;
    Stream.Write(Ver.wLength, SizeOf(Word));
    Stream.Position := StartPosition + Ver.wLength;
  end;

var
  W: Word;
  Ver: TVerHeader;
  I, CheckValue: Integer;
  StartPosition: LongInt;
begin
  StartPosition := Stream.Position;
  CheckValue := Stream.Position mod 4;
  Ver.wLength := 0; // Unknown at this time
  Ver.wValueLength := SizeOf(VS_FIXEDFILEINFO);
  Ver.wType := 0;
  // VS_VERSION_INFO.wLength
  Stream.Write(Ver.wLength, SizeOf(Word));
  // VS_VERSION_INFO.wValueLength
  Stream.Write(Ver.wValueLength, SizeOf(Word));
  // VS_VERSION_INFO.wType
  Stream.Write(Ver.wType, SizeOf(Word));
  // VS_VERSION_INFO.szKey
  for I := 1 to Length(VerInfoKey) do
  begin
    W := Ord(VerInfoKey[I]);
    Stream.Write(W, SizeOf(Word));
  end;
  // VS_VERSION_INFO.Padding1
  W := 0;
  repeat
    Stream.Write(W, SizeOf(Word))
  until Stream.Position mod 4 = CheckValue;
  // VS_VERSION_INFO.Value
  Stream.Write(FixedInfo, SizeOf(VS_FIXEDFILEINFO));
  // VS_VERSION_INFO.Padding2
  while Stream.Position mod 4 <> CheckValue do
    Stream.Write(W, SizeOf(Word));
  // VS_VERSION_INFO.Children
  WriteStringFileInfo;
  WriteVarFileInfo;
  // Update VS_VERSION_INFO.wLength
  Ver.wLength := Stream.Position - StartPosition;
  Stream.Position := StartPosition;
  Stream.Write(Ver.wLength, SizeOf(Word));
  Stream.Position := StartPosition + Ver.wLength;
end;

// Miscelaneous functions

function SplitVersion(const Ver: String): TVersion;
var
  I, J: Integer;
begin
  Result.Ver := 0;
  J := Low(Result.V);
  for I := 1 to Length(Ver) do
  begin
    if Ver[I] in ['0'..'9'] then
      Result.V[J] := Result.V[J] * 10 + Ord(Ver[I]) - Ord('0')
    else if Ver[I] = '.' then
    begin
      Inc(J);
      if J > High(Result.V) then Break;
    end;
  end;
end;

end.
